#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>

// PAM
#include <security/pam_modules.h>

// AFS

#include <afs/afsutil.h>
#include <afs/ptclient.h>
#include <afs/pterror.h>

#define MODNAME "pam_afs_groups"

int handle_req(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  int i, err, flag, sumflag;
  const char * groups = NULL, * user = NULL;
  char * ptrptr, *group, *tgroups;

  openlog(MODNAME, LOG_PID, LOG_AUTH);

  for (i=0; i<argc; i++) {
    if (strcmp(argv[i], "groups") == 0) {
      i++;
      if (i < argc) {
        groups = argv[i];
      } else {
        syslog(LOG_ERR, "no groups specified");
        return PAM_AUTH_ERR;
      }
    } else {
      syslog(LOG_WARNING, "unknown parameter %s", argv[i]);
    }
  }

  // if no groups in Pam-Config, ask Application
  if (!groups) {
    syslog(LOG_INFO, "Asking app for groups");
    struct pam_message mesg;
    const struct pam_message *mesgp = &mesg;
    struct pam_conv * conv;
    struct pam_response *resp = NULL;

    err = pam_get_item(pamh, PAM_CONV, (void*)&conv);
    if (err != PAM_SUCCESS)
      return err;

    mesg.msg_style = PAM_PROMPT_ECHO_OFF;
    mesg.msg = "GROUPS";

    err = (*(conv->conv))(1, &mesgp, &resp, conv->appdata_ptr);
    if (err != PAM_SUCCESS)
      return err;

    if (resp) {
       groups = resp->resp;
       free(resp);             /* but not resp->resp */
    }
  }

  // now ask for the user

  err=pam_get_user(pamh, &user, "");
  if (err != PAM_SUCCESS)
    return err;

  syslog(LOG_INFO, "user %s, groups %s", user, groups);

  err = pr_Initialize(0, AFSDIR_CLIENT_ETC_DIRPATH, 0); // first 0 => noauth
  if (err) {
    syslog(LOG_ERR, "Cannot connect to PT-Service");
    return PAM_AUTH_ERR;
  }

  ptrptr = malloc(strlen(groups));
  tgroups = (char *)groups;

  sumflag = 0;

  while (group = strtok_r(tgroups, ",", &ptrptr)) {
    tgroups = NULL;
    err = pr_IsAMemberOf (user, group, &flag);

    switch(err) {
      case 0:  sumflag = sumflag || flag; break;
      case PRBADARG: syslog(LOG_WARNING, "No such group %s\n", group); break;
      default: syslog(LOG_ERR, "Generic PT-Error %d\n", err); break;
    }
  }
  pr_End();

  syslog(LOG_INFO, "Result: %d", sumflag);

  return sumflag?PAM_SUCCESS:PAM_AUTH_ERR;
}

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  return handle_req(pamh, flags, argc, argv);
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  return PAM_SUCCESS;
}

PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
  return handle_req(pamh, flags, argc, argv);
}
